#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <time.h>

#include "main.h"
#include "grabber.h"
#include "util.h"

void help (FILE *out)
{
	fprintf (out,	"Usage: xkcdsw-grab [option...]\n"
			"       xkcdsw-grab -f[output dir] [option...]\n\n"
			"Options:\n"
			"	-h, --help\n"
			"		Show this message and exit.\n"
			"	-V, --version\n"
			"		Show version and exit.\n"
			"	-p <pagespec>, --pages=<pagespec>\n"
			"		Grab comics from the pages described by <pagespec> (see below).\n"
			"	-m <max>, --max-items=<max>\n"
			"		Fetch at most <max> XKCDSW comics. Default: %lu.\n"
			"	-q, --quiet\n"
			"		Suppress warnings and some error messages.\n"
			"	-v, --verbose\n"
			"		Print what the program is doing.\n"
			"	-f[output dir], --fetch[=output dir]\n"
			"		Instead of printing a list of URLs, save the files pointed to\n"
			"		by those URLs to the directory [output dir], or the current\n"
			"		directory if it isn't specified.\n"
			"		The files written will be named xkcdsw-<stamp>.<n>[.ext],\n"
			"		where <stamp> is the Unix timestamp, <n> is a number, and\n"
			"		[.ext] is '.gif', '.png' or '.jpg' for those respective\n"
			"		image formats, or is missing altogether for other formats.\n"
			"		If -f is used without -T, the title texts are also written\n"
			"		to a file, namely [output dir]/titletexts.\n"
			"	-T, --no-title-text\n"
			"		Do not print the title text of the comic, or do not write it\n"
			"		to the output file in case -f is used.\n\n"
			"pagespec:\n"
			"	The <pagespec> argument to the -p option is a comma-separated list of\n"
			"	single pages, ranges of pages, or a combination thereof.\n"
			"	Ranges are specified with a colon. The following example will grab\n"
			"	pages 1, 2, 3, 4, 5, 18, 20, 24, 25, 26, 27, 30 and 39:\n"
			"		1:5,18,20,24:27,30,39\n",
		(size_t) DEFAULT_MAX_ITEMS);
}

void version (FILE *out)
{
	fprintf (out,
			"xkcdsw-grab version 0.1\n"
			"Copyright (C) 2009 Penguin Development\n"
			"License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
			"This is free software: you are free to change and redistribute it.\n"
			"There is NO WARRANTY, to the extent permitted by law.\n\n"
			"Written by Peter 'Link' S.\n"
		);
}

int intcmp (const void *x, const void *y)
{
	int xi, yi;
	xi = *((int *) x);
	yi = *((int *) y);

	return xi - yi;
}

int parse_pages (const char *pagespec, unsigned int pagebuffer[], unsigned int flags)
{
	size_t bufptr = 0;
	memset (pagebuffer, 0, MAX_PAGES * sizeof (unsigned int));

	char *spec = strdup (pagespec);
	char *specptr = spec;
	char *specpart, *saveptr;

	int killflag = 0;

	while ((specpart = strtok_r (specptr, ",", &saveptr)) != NULL)
	{
		specptr = NULL;
		/*printf ("Debug: in parse_pages while-loop, line %i, with token `%s'\n", __LINE__, specpart);*/
		char *specpart_safe = strdup (specpart);
		char *rangespec = strchr (specpart_safe, ':');
		if (rangespec != NULL) /* Range of values */
		{
			if (strrchr (specpart_safe, ':') != rangespec || strlen (rangespec) <= 1)
			{
				if (!(flags & QUIET_FLAG))
					fprintf (stderr, "Invalid format in page spec: `%s'\n", specpart_safe);
				return 1;
			}

			char *t_start = strndup (specpart_safe, (size_t) (specpart_safe - rangespec));
			int t_pagestart = atoi (t_start);
			int t_pageend = atoi (rangespec + 1);

			free (t_start);

			if (t_pagestart < 0 || t_pageend <= t_pagestart)
			{
				if (!(flags & QUIET_FLAG))
					fprintf (stderr, "Invalid range: `%s'\n", specpart_safe);
				return 1;
			}

			unsigned int i;
			for (i = t_pagestart; i <= t_pageend; i++)
			{
				if (bufptr == MAX_PAGES - 1)
				{
					killflag = 1;
					break;
				}
				pagebuffer[bufptr] = i;
				bufptr++;
			}
		}
		else /* Single value */
		{
			if (bufptr == MAX_PAGES - 1)
				killflag = 1;
			else
			{
				int t_page = atoi (specpart_safe);

				if (t_page < 0)
				{
					if (!(flags & QUIET_FLAG))
						fprintf (stderr, "Invalid page number: `%i'.\n", t_page);
					return 1;
				}
				else if (t_page == 0)
					t_page = 1;
				pagebuffer[bufptr] = t_page;
				bufptr++;
			}
		}

		free (specpart_safe);

		if (killflag)
			break;
	}

	free (spec);

	return 0;
}

int main (int argc, char **argv)
{
	int c;
	unsigned int pages[MAX_PAGES];
	memset (pages, 0, MAX_PAGES * sizeof (unsigned int));
	unsigned int last_page, last_page_grabbed = 0;
	size_t max_items = DEFAULT_MAX_ITEMS;
	unsigned int flags = 0;
	char *output_dir = NULL;
	int killflag = -1;
	int have_pagespec = 0;

	struct xkcdsw_item *items;
	size_t item_offset = 0;
	struct grabber_handle *handle;

	while (1)
	{
		static struct option options[] =
		{
			{"help",		no_argument,		0,	'h'},
			{"version",		no_argument,		0,	'V'},
			{"pages",		required_argument,	0,	'p'},
			{"max-items",		required_argument,	0,	'm'},
			{"quiet",		no_argument,		0,	'q'},
			{"verbose",		no_argument,		0,	'v'},
			{"fetch",		optional_argument,	0,	'f'},
			{"no-title-text",	no_argument,		0,	'T'},
			{0,			0,			0,	0}
		};

		int option_index = 0;
		c = getopt_long (argc, argv, "hVp:m:qvf::T", options, &option_index);

		if (c == -1)
			break;

		switch (c)
		{
			case 'h':
				help (stdout);
				killflag = 0;
				break;
			case 'V':
				version (stdout);
				killflag = 0;
				break;
			case 'p':
				if (have_pagespec)
				{
					if (flags & VERBOSE_FLAG)
						fprintf (stderr, "Ignoring extra --pages flag; use a comma-separated list\n"
								"of pages or page ranges (using a colon) instead of multiple\n"
								"--pages specifications.\n");
					else
						fprintf (stderr, "Extra --pages ignored.\n");

					break;
				}

				if (parse_pages (optarg, pages, flags) != 0)
				{
					fprintf (stderr, "Invalid page specification `%s'.\n", optarg);
					killflag = 2;
				}
				have_pagespec = 1;

				break;
			case 'm':
			{
				int t_max = atoi (optarg);

				if (t_max < 0)
				{
					fprintf (stderr, "Notice: ignoring negative max amount.\n");
					break;
				}

				if (t_max == 0)
					max_items = DEFAULT_MAX_ITEMS;
				else
					max_items = t_max;

				break;
			}
			case 'q':
				if (flags & VERBOSE_FLAG)
				{
					fprintf (stderr, "Verbose *and* quiet? I think NOT!\n");
					killflag = 1;
					break;
				}
				flags |= QUIET_FLAG;
				break;
			case 'v':
				if (flags & QUIET_FLAG)
				{
					fprintf (stderr, "Quiet *and* verbose? I think NOT!\n");
					killflag = 1;
					break;
				}
				flags |= VERBOSE_FLAG;
				break;
			case 'f':
				flags |= FETCH_FLAG;

				if (optarg)
					output_dir = optarg;

				break;
			case 'T':
				flags |= NOTITLE_FLAG;
				break;
		}

		if (killflag != -1)
			break;
	}

	if (killflag != -1)
		return killflag;

	items = malloc (max_items * sizeof (struct xkcdsw_item));

	handle = grabber_init ();
	if (!handle)
	{
		fprintf (stderr, "Failed to initialise cURL for grabbing. :*(\n");
		return 1;
	}

	last_page = get_last_page (handle);
	if (last_page == 0)
		last_page = MAX_PAGES;

	if (!have_pagespec)
		pages[0] = last_page;

	size_t llp;
	for (llp = MAX_PAGES - 1; llp >= 0; llp--)
	{
		if (pages[llp] != 0)
			break;
	}

	qsort (pages, llp + 1, sizeof (unsigned int), intcmp);

	/*printf ("Debug: pages = {");
	size_t foo;
	for (foo = 0; foo < MAX_PAGES - 1; foo++)
	{
		printf ("%i, ", pages[foo]);
	}
	printf ("%i}\n", pages[MAX_PAGES - 1]);*/

	size_t pageptr;
	for (pageptr = 0; pageptr <= llp; pageptr++)
	{
		unsigned int page = pages[pageptr];
		if (page == 0)
			page = 1;
		else if (page > last_page)
			break;

		if (last_page_grabbed == page)
			continue;

		if (flags & VERBOSE_FLAG)
			printf ("Grabbing page %3i...", page);

		size_t grabbed = grabber_fetch_xkcdsw (handle, max_items - item_offset, &(items[item_offset]), page);

		if (flags & VERBOSE_FLAG)
			printf ("\tdone (%3lu results)\n", grabbed);

		item_offset += grabbed;
		last_page_grabbed = page;
	}

	if (flags & VERBOSE_FLAG)
		printf ("%lu comics grabbed.\n", item_offset);

	size_t i;
	time_t curtime = time (NULL);
	if (output_dir == NULL)
		output_dir = ".";

	FILE *ttfile = NULL;

	if ((flags & FETCH_FLAG) && !(flags & NOTITLE_FLAG))
	{
		char *ttfn = calloc (strlen (output_dir) + 24, 1);
		sprintf (ttfn, "%s/titletexts", output_dir);
		ttfile = fopen (ttfn, "a+");

		if (!ttfile && !(flags & QUIET_FLAG))
			fprintf (stderr, "Failed to open titletext file `%s'; title texts will not be written.\n", ttfn);

		free (ttfn);
	}

	for (i = 0; i < item_offset; i++)
	{
		struct xkcdsw_item *item = &(items[i]);

		if (flags & FETCH_FLAG)
		{
			if (flags & VERBOSE_FLAG)
				printf ("Downloading file %3lu of %3lu...", i + 1, item_offset);

			struct file_buffer buf;
			int status = download_comic (handle, &buf, item);
			if (status == 0 && (flags & VERBOSE_FLAG))
				printf ("\tdone.\n");
			else if (status != 0)
			{
				if (flags & VERBOSE_FLAG)
					printf ("\tFAILED.\n");

				if (!(flags & QUIET_FLAG))
					fprintf (stderr, "Download of comic %lu/%lu, URL %s failed.\n", i + 1, item_offset, item->url);
				continue;
			}

			char *filename_base = calloc (64, 1);
			char *filename = calloc (strlen (output_dir) + 64, 1);
			char timespec[21];
			memset (timespec, 0, 21);
			strftime (timespec, 20, "%s", gmtime (&curtime));

			sprintf (filename_base, "xkcdsw-%s.%05lu%s", timespec, i + 1, buf.ext);
			sprintf (filename, "%s%s%s", output_dir, DIR_SEP, filename_base);
			if (flags & VERBOSE_FLAG)
				printf ("Writing file %s...", filename);

			FILE *f = fopen (filename, "wb");
			if (!f)
			{
				if (flags & VERBOSE_FLAG)
					printf ("\tFAILED.\n");
				fprintf (stderr, "Failed to open file `%s' for writing.\n", filename);
				killflag = 1;
				free (filename);
				free (filename_base);
				free (buf.data);
				break;
			}

			size_t written = fwrite (buf.data, 1, buf.length, f);
			if (written != buf.length)
			{
				if (flags & VERBOSE_FLAG)
					printf ("\tFAILED.\n");
				fprintf (stderr, "Hunh? Wrote %lu bytes to file `%s'; expected %lu. Aborting.\n", written, filename, buf.length);
				killflag = 1;
				fclose (f);
				free (filename);
				free (filename_base);
				free (buf.data);
				break;
			}

			fclose (f);

			if (ttfile && strlen (item->titletext) > 0)
				fprintf (ttfile, "%s:\n\t\"%s\"\n", filename_base, item->titletext);

			free (filename);
			free (filename_base);
			if (flags & VERBOSE_FLAG)
				printf ("\tdone.\n");

			free (buf.data);
		}
		else
		{
			printf ("%3lu: %s\n", i + 1, item->url);
			if (!(flags & NOTITLE_FLAG))
			{
				if (strlen (item->titletext) > 0)
					printf ("\t\"%s\"\n\n", item->titletext);
				else
					printf ("\n");
			}
		}
	}

	if (ttfile)
		fclose (ttfile);

	free (items);

	grabber_cleanup (handle);

	if (killflag != -1)
		return killflag;

	return 0;
}
